@@ -2,19 +2,20 @@
 /*  GMime
  *  Copyright (C) 2000-2007 Jeffrey Stedfast
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  This library is free software; you can redistribute it and/or
+ *  modify it under the terms of the GNU Lesser General Public License
+ *  as published by the Free Software Foundation; either version 2.1
+ *  of the License, or (at your option) any later version.
  *
- *  This program is distributed in the hope that it will be useful,
+ *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *  Lesser General Public License for more details.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  You should have received a copy of the GNU Lesser General Public
+ *  License along with this library; if not, write to the Free
+ *  Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
+ *  02110-1301, USA.
  */
 
 
@@ -25,14 +26,19 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <locale.h>
+#include <errno.h>
 
 #ifdef HAVE_CODESET
 #include <langinfo.h>
 #endif
 
 #include "gmime-charset-map-private.h"
+#include "gmime-table-private.h"
 #include "gmime-charset.h"
+#include "gmime-iconv.h"
 
 #ifdef HAVE_ICONV_DETECT_H
 #include "iconv-detect.h"
@@ -53,6 +59,15 @@
 #endif /* USE_ICONV_DETECT */
 
 
+/**
+ * SECTION: gmime-charset
+ * @title: gmime-charset
+ * @short_description: Charset helper functions
+ * @see_also:
+ *
+ **/
+
+
 /* a useful website on charset alaises:
  * http://www.li18nux.org/subgroups/sa/locnameguide/v1.1draft/CodesetAliasTable-V11.html */
 
@@ -78,23 +93,38 @@ struct {
 	{ "iso_10646",       "UCS-2BE"    },
 	{ "iso10646",        "UCS-2BE"    },
 	
+	/* Korean charsets */
+	/* Note: according to http://www.iana.org/assignments/character-sets,
+	 * ks_c_5601-1987 should really map to ISO-2022-KR, but the EUC-KR
+	 * mapping was given to me via a native Korean user, so I'm not sure
+	 * if I should change this... perhaps they are compatable? */
 	{ "ks_c_5601-1987",  "EUC-KR"     },
 	{ "5601",            "EUC-KR"     },
 	{ "ksc-5601",        "EUC-KR"     },
 	{ "ksc-5601-1987",   "EUC-KR"     },
 	{ "ksc-5601_1987",   "EUC-KR"     },
+	{ "ks_c_5861-1992",  "EUC-KR"     },
 	{ "euckr-0",         "EUC-KR"     },
 	
-	/* FIXME: Japanese/Korean/Chinese stuff needs checking */
+	/* Chinese charsets */
 	{ "big5-0",          "BIG5"       },
 	{ "big5.eten-0",     "BIG5"       },
 	{ "big5hkscs-0",     "BIG5HKSCS"  },
-	{ "gb2312-0",        "gb2312"     },
-	{ "gb2312.1980-0",   "gb2312"     },
-	{ "euc-cn",          "gb2312"     },
+	/* Note: GBK is a superset of gb2312 (see
+	 * http://en.wikipedia.org/wiki/GBK for details), so 'upgrade'
+	 * gb2312 to GBK so that we can completely convert GBK text
+	 * that is incorrectly tagged as gb2312 to UTF-8. */
+	{ "gb2312",          "GBK"        },
+	{ "gb-2312",         "GBK"        },
+	{ "gb2312-0",        "GBK"        },
+	{ "gb2312-80",       "GBK"        },
+	{ "gb2312.1980-0",   "GBK"        },
+	/* euc-cn is an alias for gb2312 */
+	{ "euc-cn",          "GBK"        },
 	{ "gb18030-0",       "gb18030"    },
 	{ "gbk-0",           "GBK"        },
 	
+	/* Japanese charsets */
 	{ "eucjp-0",         "eucJP"  	  },  /* should this map to "EUC-JP" instead? */
 	{ "ujis-0",          "ujis"  	  },  /* we might want to map this to EUC-JP */
 	{ "jisx0208.1983-0", "SJIS"       },
@@ -129,6 +159,7 @@ static struct {
 };
 
 static GHashTable *iconv_charsets = NULL;
+static char **user_charsets = NULL;
 static char *locale_charset = NULL;
 static char *locale_lang = NULL;
 
@@ -220,6 +251,8 @@ locale_parse_lang (const char *locale)
  * Initializes the locale charset variable for later calls to
  * g_mime_locale_charset(). Only really needs to be called for non-
  * iso-8859-1 locales.
+ *
+ * Note: g_mime_init() calls this routine for you.
  **/
 void
 g_mime_charset_map_init (void)
@@ -770,237 +803,141 @@ g_mime_charset_best (const char *in, size_t inlen)
 }
 
 
-#ifdef BUILD_CHARSET_MAP
-
-#include <sys/stat.h>
-#include <errno.h>
-#include <iconv.h>
-
-static struct {
-	char *name;        /* charset name */
-	int multibyte;     /* charset type */
-	unsigned int bit;  /* assigned bit */
-} tables[] = {
-	/* These are the 8bit character sets (other than iso-8859-1,
-	 * which is special-cased) which are supported by both other
-	 * mailers and the GNOME environment. Note that the order
-	 * they're listed in is the order they'll be tried in, so put
-	 * the more-popular ones first.
-	 */
-	{ "iso-8859-2",   0, 0 },  /* Central/Eastern European */
-	{ "iso-8859-4",   0, 0 },  /* Baltic */
-	{ "koi8-r",       0, 0 },  /* Russian */
-	{ "koi8-u",       0, 0 },  /* Ukranian */
-	{ "iso-8859-5",   0, 0 },  /* Least-popular Russian encoding */
-	{ "iso-8859-7",   0, 0 },  /* Greek */
-	{ "iso-8859-8",   0, 0 },  /* Hebrew; Visual */
-	{ "iso-8859-9",   0, 0 },  /* Turkish */
-	{ "iso-8859-13",  0, 0 },  /* Baltic again */
-	{ "iso-8859-15",  0, 0 },  /* New-and-improved iso-8859-1, but most
-				    * programs that support this support UTF8
-				    */
-	{ "windows-1251", 0, 0 },  /* Russian */
-	
-	/* These are the multibyte character sets which are commonly
-	 * supported by other mail clients. Note: order for multibyte
-	 * charsets does not affect priority unlike the 8bit charsets
-	 * listed above.
-	 */
-	{ "iso-2022-jp",  1, 0 },  /* Japanese designed for use over the Net */
-	{ "Shift-JIS",    1, 0 },  /* Japanese as used by Windows and MacOS systems */
-	{ "euc-jp",       1, 0 },  /* Japanese traditionally used on Unix systems */
-	{ "euc-kr",       1, 0 },  /* Korean */
-	{ "iso-2022-kr",  1, 0 },  /* Korean (less popular than euc-kr) */
-	{ "gb2312",       1, 0 },  /* Simplified Chinese */
-	{ "Big5",         1, 0 },  /* Traditional Chinese */
-	{ "euc-tw",       1, 0 },
-	{ NULL,           0, 0 }
-};
-
-unsigned int encoding_map[256 * 256];
-
-#if G_BYTE_ORDER == G_BIG_ENDIAN
-#define UCS "UCS-4BE"
-#else
-#define UCS "UCS-4LE"
-#endif
-
-int main (int argc, char **argv)
+/**
+ * g_mime_charset_can_encode:
+ * @mask: a #GMimeCharset mask
+ * @charset: a charset
+ * @text: utf-8 text to check
+ * @len: length of @text
+ *
+ * Check to see if the UTF-8 @text will fit safely within @charset.
+ *
+ * Returns %TRUE if it is safe to encode @text into @charset or %FALSE
+ * otherwise.
+ **/
+gboolean
+g_mime_charset_can_encode (GMimeCharset *mask, const char *charset, const char *text, size_t len)
 {
-	GHashTable *table_hash;
-	size_t inleft, outleft;
-	char *inbuf, *outbuf;
-	guint32 out[128], c;
-	unsigned int bit = 0x01;
-	char in[128];
-	int i, j, k;
-	int bytes;
+	const unsigned char *inptr = (const unsigned char *) text;
+	const unsigned char *inend = inptr + len;
+	size_t inleft, outleft, rc;
+	const char *inbuf = text;
+	char out[256], *outbuf;
+	const char *iconv_name;
 	iconv_t cd;
+	guint i;
 	
-	/* dont count the terminator */
-	bytes = ((sizeof (tables) / sizeof (tables[0])) + 7 - 1) / 8;
+	if (len == 0)
+		return TRUE;
 	
-	for (i = 0; i < 128; i++)
-		in[i] = i + 128;
+	if (mask->level == 0 && (!charset || !g_ascii_strcasecmp (charset, "us-ascii"))) {
+		/* simple US-ASCII case - is this scan even necessary? */
+		while (inptr < inend && is_ascii (*inptr))
+			inptr++;
+		
+		if (inptr == inend)
+			return TRUE;
+		
+		return FALSE;
+	}
 	
-	for (j = 0; tables[j].name && !tables[j].multibyte; j++) {
-		cd = iconv_open (UCS, tables[j].name);
-		inbuf = in;
-		outbuf = (char *)(out);
-		inleft = sizeof (in);
+	if (!g_ascii_strcasecmp (charset, "utf-8")) {
+		/* we can encode anything in utf-8 */
+		return TRUE;
+	}
+	
+	charset = g_mime_charset_iconv_name (charset);
+	
+	if (mask->level == 1)
+		return !g_ascii_strcasecmp (charset, "iso-8859-1");
+	
+	/* check if this is a charset that we have precalculated masking for */
+	for (i = 0; i < G_N_ELEMENTS (charinfo); i++) {
+		iconv_name = g_mime_charset_iconv_name (charinfo[i].name);
+		if (charset == iconv_name)
+			break;
+	}
+	
+	if (i < G_N_ELEMENTS (charinfo)) {
+		/* indeed we do... */
+		return (charinfo[i].bit & mask->mask);
+	}
+	
+	/* down to the nitty gritty slow and painful way... */
+	if ((cd = g_mime_iconv_open (charset, "UTF-8")) == (iconv_t) -1)
+		return FALSE;
+	
+	inleft = len;
+	
+	do {
 		outleft = sizeof (out);
-		while (iconv (cd, &inbuf, &inleft, &outbuf, &outleft) == -1) {
-			if (errno == EILSEQ) {
-				inbuf++;
-				inleft--;
-			} else {
-				g_warning ("iconv (%s->UCS4, ..., %d, ..., %d): %s",
-					   tables[j].name, inleft, outleft,
-					   g_strerror (errno));
-				exit (1);
-			}
-		}
-		iconv_close (cd);
+		outbuf = out;
+		errno = 0;
 		
-		for (i = 0; i < 128 - outleft / 4; i++) {
-			encoding_map[i] |= bit;
-			encoding_map[out[i]] |= bit;
-		}
+		rc = iconv (cd, (char **) &inbuf, &inleft, &outbuf, &outleft);
+		if (rc == (size_t) -1 && errno != E2BIG)
+			break;
+	} while (inleft > 0);
+	
+	if (inleft == 0) {
+		outleft = sizeof (out);
+		outbuf = out;
+		errno = 0;
 		
-		tables[j].bit = bit;
-		bit <<= 1;
+		rc = iconv (cd, NULL, NULL, &outbuf, &outleft);
 	}
 	
-	/* Mutibyte tables */
-	for ( ; tables[j].name && tables[j].multibyte; j++) {
-		cd = iconv_open (tables[j].name, UCS);
-		if (cd == (iconv_t) -1)
-			continue;
-		
-		for (c = 128, i = 0; c < 65535 && i < 65535; c++) {
-			inbuf = (char *) &c;
-			inleft = sizeof (c);
-			outbuf = in;
-			outleft = sizeof (in);
-			
-			if (iconv (cd, &inbuf, &inleft, &outbuf, &outleft) != (size_t) -1) {
-				/* this is a legal character in charset table[j].name */
-				iconv (cd, NULL, NULL, &outbuf, &outleft);
-				encoding_map[i++] |= bit;
-				encoding_map[c] |= bit;
-			} else {
-				/* reset the iconv descriptor */
-				iconv (cd, NULL, NULL, NULL, NULL);
-			}
-		}
-		
-		iconv_close (cd);
-		
-		tables[j].bit = bit;
-		bit <<= 1;
-	}
+	g_mime_iconv_close (cd);
 	
-	printf ("/* This file is automatically generated: DO NOT EDIT */\n\n");
-	
-	/* FIXME: we can condense better than what my quick hack does,
-	   but it'd be more work and I'm not sure if it's worth it or
-	   not. Currently I'm just making it so that tables that
-	   contain all of the same values will only ever be
-	   one-of-a-kind by making duplicates into macro aliases for
-	   the original */
-	
-	table_hash = g_hash_table_new (g_int_hash, g_int_equal);
-	
-	for (i = 0; i < 256; i++) {
-		/* first, do we need this block? */
-		for (k = 0; k < bytes; k++) {
-			int first = encoding_map[i * 256] & (0xff << (k * 8));
-			int same = TRUE;
-			int dump = FALSE;
-			
-			for (j = 0; j < 256; j++) {
-				same = same && (encoding_map[i * 256 + j] & (0xff << (k * 8))) == first;
-				if ((encoding_map[i * 256 + j] & (0xff << (k * 8))) != 0)
-					dump = TRUE;
-			}
-			
-			if (dump) {
-				if (same) {
-					/* this table is aliasable */
-					char *table_name;
-					
-					if ((table_name = g_hash_table_lookup (table_hash, &first))) {
-						/* we've already written out a table with the exact same
-						   values so we can just alias it with a macro. */
-						printf ("#define m%02x%x %s\n\n", i, k, table_name);
-						continue;
-					} else {
-						table_name = g_strdup_printf ("m%02x%x", i, k);
-						g_hash_table_insert (table_hash, &first, table_name);
-					}
-				}
-				
-				/* yes, dump it */
-				printf ("static unsigned char m%02x%x[256] = {\n\t", i, k);
-				for (j = 0; j < 256; j++) {
-					printf ("0x%02x, ", (encoding_map[i * 256 + j] >> (k * 8)) & 0xff);
-					if (((j + 1) & 7) == 0 && j < 255)
-						printf ("\n\t");
-				}
-				printf ("\n};\n\n");
-			}
-		}
-	}
-	
-	printf ("struct {\n");
-	for (k = 0; k < bytes; k++) {
-		printf ("\tunsigned char *bits%d;\n", k);
-	}
-	
-	printf ("} charmap[256] = {\n\t");
-	for (i = 0; i < 256; i++) {
-		/* first, do we need this block? */
-		printf ("{ ");
-		for (k = 0; k < bytes; k++) {
-			for (j = 0; j < 256; j++) {
-				if ((encoding_map[i * 256 + j] & (0xff << (k * 8))) != 0)
-					break;
-			}
-			
-			if (j < 256) {
-				printf ("m%02x%x, ", i, k);
-			} else {
-				printf ("0, ");
-			}
-		}
-		
-		printf ("}, ");
-		if (((i + 1) & 7) == 0 && i < 255)
-			printf ("\n\t");
+	return rc != (size_t) -1;
+}
+
+
+/**
+ * g_mime_set_user_charsets:
+ * @charsets: an array of user-preferred charsets
+ *
+ * Set a list of charsets for GMime to use as a hint for encoding and
+ * decoding headers. The charset list should be in order of preference
+ * (e.g. most preferred first, least preferred last).
+ **/
+void
+g_mime_set_user_charsets (const char **charsets)
+{
+	int i, n = 0;
+	
+	if (user_charsets)
+		g_strfreev (user_charsets);
+	
+	if (charsets == NULL) {
+		user_charsets = NULL;
+		return;
 	}
-	printf ("\n};\n\n");
 	
-	printf ("struct {\n\tconst char *name;\n\tunsigned int bit;\n} charinfo[] = {\n");
-	for (j = 0; tables[j].name; j++) {
-		printf ("\t{ \"%s\", 0x%04x },\n", tables[j].name, tables[j].bit);
-	}
-	printf ("};\n\n");
+	while (charsets[n] != NULL)
+		n++;
 	
-	printf("#define charset_mask(x) \\\n");
-	for (k = 0; k < bytes; k++) {
-		if (k != 0)
-			printf ("\t| ");
-		else
-			printf ("\t");
-		
-		printf ("(charmap[(x) >> 8].bits%d ? charmap[(x) >> 8].bits%d[(x) & 0xff] << %d : 0)",
-			k, k, k * 8);
-		
-		if (k < bytes - 1)
-			printf ("\t\\\n");
+	if (n == 0) {
+		user_charsets = NULL;
+		return;
 	}
-	printf ("\n\n");
 	
-	return 0;
+	user_charsets = g_malloc (sizeof (char *) * (n + 1));
+	for (i = 0; i <= n; i++)
+		user_charsets[i] = (char *) charsets[i];
+}
+
+
+/**
+ * g_mime_user_charsets:
+ *
+ * Get the list of user-preferred charsets set with
+ * g_mime_set_user_charsets().
+ *
+ * Returns an array of user-set charsets or %NULL if none set.
+ **/
+const char **
+g_mime_user_charsets (void)
+{
+	return (const char **) user_charsets;
 }
-#endif /* BUILD_CHARSET_MAP */
