@@ -2,19 +2,20 @@
 /*  GMime
  *  Copyright (C) 2000-2007 Jeffrey Stedfast
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  This library is free software; you can redistribute it and/or
+ *  modify it under the terms of the GNU Lesser General Public License
+ *  as published by the Free Software Foundation; either version 2.1
+ *  of the License, or (at your option) any later version.
  *
- *  This program is distributed in the hope that it will be useful,
+ *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *  Lesser General Public License for more details.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  You should have received a copy of the GNU Lesser General Public
+ *  License along with this library; if not, write to the Free
+ *  Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
+ *  02110-1301, USA.
  */
 
 
@@ -56,6 +57,18 @@
 
 #define d(x)
 
+
+/**
+ * SECTION: gmime-utils
+ * @title: gmime-utils
+ * @short_description: MIME utility functions
+ * @see_also:
+ *
+ * Utility functions to parse, encode and decode various MIME tokens
+ * and encodings.
+ **/
+
+
 #define GMIME_UUENCODE_CHAR(c) ((c) ? (c) + ' ' : '`')
 #define	GMIME_UUDECODE_CHAR(c) (((c) - ' ') & 077)
 
@@ -644,7 +657,7 @@ parse_broken_date (struct _date_token *tokens, int *tzone)
 					goto next;
 				} else if (!tm.tm_year) {
 					if ((n = get_year (token->start, token->len)) != -1) {
-						printf ("2-digit year; ");
+						d(printf ("2-digit year; "));
 						tm.tm_year = n - 1900;
 					}
 					goto next;
@@ -1056,9 +1069,10 @@ is_rfc2047_token (const char *inptr, size_t len)
 	if (len < 8 || strncmp (inptr, "=?", 2) != 0 || strncmp (inptr + len - 2, "?=", 2) != 0)
 		return FALSE;
 	
-	inptr += 3;
-	len -= 3;
+	inptr += 2;
+	len -= 2;
 	
+	/* skip past the charset */
 	while (*inptr != '?' && len > 0) {
 		inptr++;
 		len--;
@@ -1399,74 +1413,172 @@ g_mime_utils_best_encoding (const unsigned char *text, size_t len)
 }
 
 
-static char *
-decode_8bit (const char *text, size_t len)
+/**
+ * charset_convert:
+ * @cd: iconv converter
+ * @inbuf: input text buffer to convert
+ * @inleft: length of the input buffer
+ * @outp: pointer to output buffer
+ * @outlenp: pointer to output buffer length
+ * @ninval: the number of invalid bytes in @inbuf
+ *
+ * Converts the input buffer from one charset to another using the
+ * @cd. On completion, @outp will point to the output buffer
+ * containing the converted text (nul-terminated), @outlenp will be
+ * the size of the @outp buffer (note: not the strlen() of @outp) and
+ * @ninval will contain the number of bytes which could not be
+ * converted.
+ *
+ * Bytes which cannot be converted from @inbuf will appear as '?'
+ * characters in the output buffer.
+ *
+ * If *@outp is non-NULL, then it is assumed that it points to a
+ * pre-allocated buffer of length *@outlenp. This is done so that the
+ * same output buffer can be reused multiple times.
+ *
+ * Returns the string length of the output buffer.
+ **/
+static size_t
+charset_convert (iconv_t cd, const char *inbuf, size_t inleft, char **outp, size_t *outlenp, size_t *ninval)
 {
-	const char *fallback_charsets[3] = { "UTF-8", NULL, NULL };
-	size_t inleft, outleft, outlen, rc, min, n;
-	const char **charsets, *best;
-	char *out, *outbuf;
-	const char *inbuf;
-	iconv_t cd;
-	int i;
+	size_t outlen, outleft, rc, n = 0;
+	char *outbuf, *out;
 	
-	if (!(charsets = g_mime_user_charsets ())) {
-		inbuf = g_mime_locale_charset ();
-		if (g_ascii_strcasecmp (inbuf, "UTF-8") != 0)
-			fallback_charsets[1] = inbuf;
-		
-		charsets = fallback_charsets;
+	if (*outp == NULL) {
+		outleft = outlen = (inleft * 2) + 16;
+		outbuf = out = g_malloc (outlen + 1);
+	} else {
+		outleft = outlen = *outlenp;
+		outbuf = out = *outp;
 	}
 	
+	do {
+		rc = iconv (cd, (char **) &inbuf, &inleft, &outbuf, &outleft);
+		if (rc == (size_t) -1) {
+			if (errno == EINVAL) {
+				/* incomplete sequence at the end of the input buffer */
+				n += inleft;
+				break;
+			}
+			
+			if (errno == E2BIG) {
+				/* need to grow the output buffer */
+				outlen += (inleft * 2) + 16;
+				rc = (size_t) (outbuf - out);
+				out = g_realloc (out, outlen + 1);
+				outleft = outlen - rc;
+				outbuf = out + rc;
+			} else {
+				/* invalid byte(-sequence) in the input buffer */
+				*outbuf++ = '?';
+				outleft--;
+				inleft--;
+				inbuf++;
+				n++;
+			}
+		}
+	} while (inleft > 0);
+	
+	iconv (cd, NULL, NULL, &outbuf, &outleft);
+	*outbuf++ = '\0';
+	
+	*outlenp = outlen;
+	*outp = out;
+	*ninval = n;
+	
+	return (outbuf - out);
+}
+
+
+#define USER_CHARSETS_INCLUDE_UTF8    (1 << 0)
+#define USER_CHARSETS_INCLUDE_LOCALE  (1 << 1)
+
+
+/**
+ * g_mime_utils_decode_8bit:
+ * @text: input text in unknown 8bit/multibyte character set
+ * @len: input text length
+ *
+ * Attempts to convert text in an unknown 8bit/multibyte charset into
+ * UTF-8 by finding the charset which will convert the most bytes into
+ * valid UTF-8 characters as possible. If no exact match can be found,
+ * it will choose the best match and convert invalid byte sequences
+ * into question-marks (?) in the returned string buffer.
+ *
+ * Returns a UTF-8 string representation of @text.
+ **/
+char *
+g_mime_utils_decode_8bit (const char *text, size_t len)
+{
+	const char **charsets, **user_charsets, *locale, *best;
+	size_t outleft, outlen, min, ninval;
+	unsigned int included = 0;
+	char *out, *outbuf;
+	iconv_t cd;
+	int i = 0;
+	
+	g_return_val_if_fail (text != NULL, NULL);
+	
+	locale = g_mime_locale_charset ();
+	if (locale && !g_ascii_strcasecmp (locale, "UTF-8"))
+		included |= USER_CHARSETS_INCLUDE_LOCALE;
+	
+	if ((user_charsets = g_mime_user_charsets ())) {
+		while (user_charsets[i])
+			i++;
+	}
+	
+	charsets = g_alloca (sizeof (char *) * (i + 3));
+	i = 0;
+	
+	if (user_charsets) {
+		while (user_charsets[i]) {
+			/* keep a record of whether or not the user-supplied
+			 * charsets include UTF-8 and/or the default fallback
+			 * charset so that we avoid doubling our efforts for
+			 * these 2 charsets. We could have used a hash table
+			 * to keep track of unique charsets, but we can
+			 * (hopefully) assume that user_charsets is a unique
+			 * list of charsets with no duplicates. */
+			if (!g_ascii_strcasecmp (user_charsets[i], "UTF-8"))
+				included |= USER_CHARSETS_INCLUDE_UTF8;
+			
+			if (locale && !g_ascii_strcasecmp (user_charsets[i], locale))
+				included |= USER_CHARSETS_INCLUDE_LOCALE;
+			
+			charsets[i] = user_charsets[i];
+			i++;
+		}
+	}
+	
+	if (!(included & USER_CHARSETS_INCLUDE_UTF8))
+		charsets[i++] = "UTF-8";
+	
+	if (!(included & USER_CHARSETS_INCLUDE_LOCALE))
+		charsets[i++] = locale;
+	
+	charsets[i] = NULL;
+	
 	min = len;
 	best = charsets[0];
 	
-	outlen = (len * 2) + 16; 
-	out = g_malloc (outlen + 1);
+	outleft = (len * 2) + 16;
+	out = g_malloc (outleft + 1);
 	
 	for (i = 0; charsets[i]; i++) {
 		if ((cd = g_mime_iconv_open ("UTF-8", charsets[i])) == (iconv_t) -1)
 			continue;
 		
-		outleft = outlen;
-		outbuf = out;
-		inleft = len;
-		inbuf = text;
-		n = 0;
-		
-		do {
-			rc = iconv (cd, (char **) &inbuf, &inleft, &outbuf, &outleft);
-			if (rc == (size_t) -1) {
-				if (errno == EINVAL) {
-					/* incomplete sequence at the end of the input buffer */
-					n += inleft;
-					break;
-				}
-				
-				if (errno == E2BIG) {
-					outlen += (inleft * 2) + 16;
-					rc = (size_t) (outbuf - out);
-					out = g_realloc (out, outlen + 1);
-					outbuf = out + rc;
-				} else {
-					inleft--;
-					inbuf++;
-					n++;
-				}
-			}
-		} while (inleft > 0);
-		
-		rc = iconv (cd, NULL, NULL, &outbuf, &outleft);
-		*outbuf = '\0';
+		outlen = charset_convert (cd, text, len, &out, &outleft, &ninval);
 		
 		g_mime_iconv_close (cd);
 		
-		if (rc != (size_t) -1 && n == 0)
-			return out;
+		if (ninval == 0)
+			return g_realloc (out, outlen + 1);
 		
-		if (n < min) {
+		if (ninval < min) {
 			best = charsets[i];
-			min = n;
+			min = ninval;
 		}
 	}
 	
@@ -1480,8 +1592,7 @@ decode_8bit (const char *text, size_t len)
 		 * is replace the 8bit garbage and pray */
 		register const char *inptr = text;
 		const char *inend = inptr + len;
-		
-		outbuf = out;
+		char *outbuf = out;
 		
 		while (inptr < inend) {
 			if (is_ascii (*inptr))
@@ -1490,46 +1601,19 @@ decode_8bit (const char *text, size_t len)
 				*outbuf++ = '?';
 		}
 		
-		*outbuf = '\0';
+		*outbuf++ = '\0';
 		
-		return out;
+		return g_realloc (out, outbuf - out);
 	}
 	
-	outleft = outlen;
-	outbuf = out;
-	inleft = len;
-	inbuf = text;
-	
-	do {
-		rc = iconv (cd, (char **) &inbuf, &inleft, &outbuf, &outleft);
-		if (rc == (size_t) -1) {
-			if (errno == EINVAL) {
-				/* incomplete sequence at the end of the input buffer */
-				break;
-			}
-			
-			if (errno == E2BIG) {
-				rc = outbuf - out;
-				outlen += inleft * 2 + 16;
-				out = g_realloc (out, outlen + 1);
-				outbuf = out + rc;
-			} else {
-				*outbuf++ = '?';
-				outleft--;
-				inleft--;
-				inbuf++;
-			}
-		}
-	} while (inleft > 0);
-	
-	iconv (cd, NULL, NULL, &outbuf, &outleft);
-	*outbuf = '\0';
+	outlen = charset_convert (cd, text, len, &out, &outleft, &ninval);
 	
 	g_mime_iconv_close (cd);
 	
-	return out;
+	return g_realloc (out, outlen + 1);
 }
 
+
 /* this decodes rfc2047's version of quoted-printable */
 static ssize_t
 quoted_decode (const unsigned char *in, size_t len, unsigned char *out)
@@ -1576,14 +1660,15 @@ rfc2047_decode_word (const char *in, size_t inlen)
 	const unsigned char *inend = instart + inlen - 2;
 	unsigned char *decoded;
 	const char *charset;
+	size_t len, ninval;
 	char *charenc, *p;
 	guint32 save = 0;
 	ssize_t declen;
 	int state = 0;
-	size_t len;
 	iconv_t cd;
 	char *buf;
 	
+	/* skip over the charset */
 	if (!(inptr = memchr (inptr, '?', inend - inptr)) || inptr[2] != '?')
 		return NULL;
 	
@@ -1642,23 +1727,29 @@ rfc2047_decode_word (const char *in, size_t inlen)
 		return g_strndup ((char *) decoded, declen);
 	}
 	
-	if ((cd = g_mime_iconv_open ("UTF-8", charset)) == (iconv_t) -1) {
+	if (!charset[0] || (cd = g_mime_iconv_open ("UTF-8", charset)) == (iconv_t) -1) {
 		w(g_warning ("Cannot convert from %s to UTF-8, header display may "
-			     "be corrupt: %s", charset, g_strerror (errno)));
+			     "be corrupt: %s", charset[0] ? charset : "unspecified charset",
+			     g_strerror (errno)));
 		
-		return decode_8bit (decoded, declen);
+		return g_mime_utils_decode_8bit ((char *) decoded, declen);
 	}
 	
-	buf = g_mime_iconv_strndup (cd, (char *) decoded, declen);
+	len = declen;
+	buf = g_malloc (len + 1);
+	
+	charset_convert (cd, decoded, declen, &buf, &len, &ninval);
+	
 	g_mime_iconv_close (cd);
 	
-	if (buf != NULL)
-		return buf;
+#if w(!)0
+	if (ninval > 0) {
+		g_warning ("Failed to completely convert \"%.*s\" to UTF-8, display may be "
+			   "corrupt: %s", declen, decoded, g_strerror (errno));
+	}
+#endif
 	
-	w(g_warning ("Failed to convert \"%.*s\" to UTF-8, display may be "
-		     "corrupt: %s", declen, decoded, g_strerror (errno)));
-	
-	return decode_8bit (decoded, declen);
+	return buf;
 }
 
 
@@ -1702,10 +1793,46 @@ g_mime_utils_header_decode_text (const char *in)
 			text = inptr;
 			ascii = TRUE;
 			
+#ifdef ENABLE_RFC2047_WORKAROUNDS
+			if (!strncmp (inptr, "=?", 2)) {
+				inptr += 2;
+				
+				/* skip past the charset (if one is even declared, sigh) */
+				while (*inptr && *inptr != '?') {
+					ascii = ascii && is_ascii (*inptr);
+					inptr++;
+				}
+				
+				/* sanity check encoding type */
+				if (inptr[0] != '?' || !strchr ("BbQq", inptr[1]) || inptr[2] != '?')
+					goto non_rfc2047;
+				
+				inptr += 3;
+				
+				/* find the end of the rfc2047 encoded word token */
+				while (*inptr && strncmp (inptr, "?=", 2) != 0) {
+					ascii = ascii && is_ascii (*inptr);
+					inptr++;
+				}
+				
+				if (!strncmp (inptr, "?=", 2))
+					inptr += 2;
+			} else {
+			non_rfc2047:
+				/* stop if we encounter a possible rfc2047 encoded
+				 * token even if it's inside another word, sigh. */
+				while (*inptr && !is_lwsp (*inptr) &&
+				       strncmp (inptr, "=?", 2) != 0) {
+					ascii = ascii && is_ascii (*inptr);
+					inptr++;
+				}
+			}
+#else
 			while (*inptr && !is_lwsp (*inptr)) {
 				ascii = ascii && is_ascii (*inptr);
 				inptr++;
 			}
+#endif /* ENABLE_RFC2047_WORKAROUNDS */
 			
 			n = (size_t) (inptr - text);
 			if (is_rfc2047_encoded_word (text, n)) {
@@ -1731,7 +1858,7 @@ g_mime_utils_header_decode_text (const char *in)
 				/* append word token */
 				if (!ascii) {
 					/* *sigh* I hate broken mailers... */
-					decoded = decode_8bit (text, n);
+					decoded = g_mime_utils_decode_8bit (text, n);
 					g_string_append (out, decoded);
 					g_free (decoded);
 				} else {
@@ -1830,7 +1957,7 @@ g_mime_utils_header_decode_phrase (const char *in)
 			
 			if (!ascii) {
 				/* *sigh* I hate broken mailers... */
-				decoded = decode_8bit (text, n);
+				decoded = g_mime_utils_decode_8bit (text, n);
 				g_string_append (out, decoded);
 				g_free (decoded);
 			} else {
@@ -1861,7 +1988,7 @@ quoted_encode (const char *in, size_t len, unsigned char *out, gushort safemask)
 		c = *inptr++;
 		if (c == ' ') {
 			*outptr++ = '_';
-		} else if (gmime_special_table[c] & safemask) {
+		} else if (c != '_' && gmime_special_table[c] & safemask) {
 			*outptr++ = c;
 		} else {
 			*outptr++ = '=';
@@ -1960,21 +2087,6 @@ struct _rfc822_word {
 	int encoding;
 };
 
-static gboolean
-word_types_compatable (enum _rfc822_word_t type1, enum _rfc822_word_t type2)
-{
-	switch (type1) {
-	case WORD_ATOM:
-		return type2 != WORD_ATOM;
-	case WORD_QSTRING:
-		return type2 != WORD_2047;
-	case WORD_2047:
-		return type2 == WORD_2047;
-	default:
-		return FALSE;
-	}
-}
-
 /* okay, so 'unstructured text' fields don't actually contain 'word'
  * tokens, but we can group stuff similarly... */
 static struct _rfc822_word *
@@ -2046,8 +2158,10 @@ rfc2047_encode_get_rfc822_words (const char *in, gboolean phrase)
 				tail = word;
 				count = 0;
 				
+				/* Note: don't reset 'type' as it
+				 * needs to be preserved when breaking
+				 * long words */
 				start = inptr;
-				type = WORD_ATOM;
 				encoding = 0;
 			}
 		}
@@ -2067,54 +2181,116 @@ rfc2047_encode_get_rfc822_words (const char *in, gboolean phrase)
 		tail = word;
 	}
 	
+#if d(!)0
+	printf ("rfc822 word tokens:\n");
+	word = words;
+	while (word) {
+		printf ("\t'%.*s'; type=%d, encoding=%d\n",
+			word->end - word->start, word->start,
+			word->type, word->encoding);
+		
+		word = word->next;
+	}
+#endif
+	
 	return words;
 }
 
 #define MERGED_WORD_LT_FOLDLEN(wlen, type) ((type) == WORD_2047 ? (wlen) < GMIME_FOLD_PREENCODED : (wlen) < (GMIME_FOLD_LEN - 8))
 
 static gboolean
+should_merge_words (struct _rfc822_word *word, struct _rfc822_word *next)
+{
+	switch (word->type) {
+	case WORD_ATOM:
+		if (next->type == WORD_2047)
+			return FALSE;
+		
+		return (MERGED_WORD_LT_FOLDLEN (next->end - word->start, next->type));
+	case WORD_QSTRING:
+		/* avoid merging with words that need to be rfc2047 encoded */
+		if (next->type == WORD_2047)
+			return FALSE;
+		
+		return (MERGED_WORD_LT_FOLDLEN (next->end - word->start, WORD_QSTRING));
+	case WORD_2047:
+		if (next->type == WORD_ATOM) {
+			/* whether we merge or not is dependent upon:
+			 * 1. the number of atoms in a row after 'word'
+			 * 2. if there is another encword after the string of atoms.
+			 */
+			int natoms = 0;
+			
+			while (next && next->type == WORD_ATOM) {
+				next = next->next;
+				natoms++;
+			}
+			
+			/* if all the words after the encword are atoms, don't merge */
+			if (!next || natoms > 3)
+				return FALSE;
+		}
+		
+		/* avoid merging with qstrings */
+		if (next->type == WORD_QSTRING)
+			return FALSE;
+		
+		return (MERGED_WORD_LT_FOLDLEN (next->end - word->start, WORD_2047));
+	default:
+		return FALSE;
+	}
+}
+
+static void
 rfc2047_encode_merge_rfc822_words (struct _rfc822_word **wordsp)
 {
 	struct _rfc822_word *word, *next, *words = *wordsp;
-	gboolean merged = FALSE;
+	gboolean rfc2047_words = FALSE;
 	
-	/* scan the list, checking for words of similar types that can be merged */
+	/* first pass: merge qstrings with adjacent qstrings and encwords with adjacent encwords */
 	word = words;
-	while (word) {
+	while (word && word->next) {
 		next = word->next;
 		
-		while (next) {
-			/* merge nodes of the same type AND we are not creating too long a string */
-			if (word_types_compatable (word->type, next->type)) {
-				if (MERGED_WORD_LT_FOLDLEN (next->end - word->start, MAX (word->type, next->type))) {
-					/* the resulting word type is the MAX of the 2 types */
-					word->type = MAX (word->type, next->type);
-					
-					word->end = next->end;
-					word->next = next->next;
-					
-					g_free (next);
-					
-					next = word->next;
-					
-					merged = TRUE;
-				} else {
-					/* if it is going to be too long, make sure we include the
-					   separating whitespace */
-					word->end = next->start;
-					break;
-				}
-			} else {
-				break;
-			}
+		if (word->type != WORD_ATOM && word->type == next->type &&
+		    MERGED_WORD_LT_FOLDLEN (next->end - word->start, word->type)) {
+			/* merge the words */
+			word->encoding = MAX (word->encoding, next->encoding);
+			
+			word->end = next->end;
+			word->next = next->next;
+			
+			g_free (next);
+			
+			next = word;
 		}
 		
-		word = word->next;
+		word = next;
+	}
+	
+	/* second pass: now merge atoms with the other words */
+	word = words;
+	while (word && word->next) {
+		next = word->next;
+		
+		if (should_merge_words (word, next)) {
+			/* the resulting word type is the MAX of the 2 types */
+			word->type = MAX (word->type, next->type);
+			
+			word->encoding = MAX (word->encoding, next->encoding);
+			
+			word->end = next->end;
+			word->next = next->next;
+			
+			g_free (next);
+			
+			continue;
+		}
+		
+		word = next;
 	}
 	
 	*wordsp = words;
-	
-	return merged;
 }
 
 static void
@@ -2155,8 +2331,7 @@ rfc2047_encode (const char *in, gushort safemask)
 	if (!(words = rfc2047_encode_get_rfc822_words (in, safemask & IS_PSAFE)))
 		return g_strdup (in);
 	
-	while (rfc2047_encode_merge_rfc822_words (&words))
-		;
+	rfc2047_encode_merge_rfc822_words (&words);
 	
 	charsets = g_mime_user_charsets ();
 	
@@ -2194,9 +2369,14 @@ rfc2047_encode (const char *in, gushort safemask)
 				start = word->start;
 			}
 			
-			if (word->encoding == 1) {
+			switch (word->encoding) {
+			case 0: /* us-ascii */
+				rfc2047_encode_word (out, start, len, "us-ascii", safemask);
+				break;
+			case 1: /* iso-8859-1 */
 				rfc2047_encode_word (out, start, len, "iso-8859-1", safemask);
-			} else {
+				break;
+			default:
 				charset = NULL;
 				g_mime_charset_init (&mask);
 				g_mime_charset_step (&mask, start, len);
@@ -2212,7 +2392,9 @@ rfc2047_encode (const char *in, gushort safemask)
 					charset = g_mime_charset_best_name (&mask);
 				
 				rfc2047_encode_word (out, start, len, charset, safemask);
+				break;
 			}
+			
 			break;
 		}
 		
