@@ -26,6 +26,15 @@
 
 G_BEGIN_DECLS
 
+
+/**
+ * InternetAddressType:
+ * @INTERNET_ADDRESS_NONE: No type.
+ * @INTERNET_ADDRESS_NAME: A typical internet address type.
+ * @INTERNET_ADDRESS_GROUP: An rfc822 group type address.
+ *
+ * The type of #InternetAddress.
+ **/
 typedef enum {
 	INTERNET_ADDRESS_NONE,
 	INTERNET_ADDRESS_NAME,
@@ -35,6 +44,15 @@ typedef enum {
 typedef struct _InternetAddress InternetAddress;
 typedef struct _InternetAddressList InternetAddressList;
 
+
+/**
+ * InternetAddress:
+ * @type: The type of internet address.
+ * @refcount: The reference count.
+ * @name: The name component of the internet address.
+ *
+ * A structure representing an rfc822 address.
+ **/
 struct _InternetAddress {
 	InternetAddressType type;
 	unsigned int refcount;
@@ -45,11 +63,20 @@ struct _InternetAddress {
 	} value;
 };
 
+
+/**
+ * InternetAddressList:
+ * @next: Pointer to the next item in the list.
+ * @address: The #InternetAddress.
+ *
+ * A list of #InternetAddress structures.
+ **/
 struct _InternetAddressList {
 	struct _InternetAddressList *next;
 	InternetAddress *address;
 };
 
+
 InternetAddress *internet_address_new (void);
 InternetAddress *internet_address_new_name (const char *name, const char *addr);
 InternetAddress *internet_address_new_group (const char *name);
@@ -80,6 +107,8 @@ InternetAddressList *internet_address_parse_string (const char *string);
 char *internet_address_to_string (const InternetAddress *ia, gboolean encode);
 char *internet_address_list_to_string (const InternetAddressList *list, gboolean encode);
 
+void internet_address_list_writer (const InternetAddressList *list, GString *str);
+
 G_END_DECLS
 
 #endif /* __INTERNET_ADDRESS_H__ */
