@@ -18,6 +18,10 @@
  */
 
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -123,8 +127,11 @@ test_parser (GMimeStream *stream)
 	{
 		GMimeStream *stream;
 		
+		g_mime_header_set_raw (GMIME_OBJECT (message)->headers, NULL);
+		
 		fprintf (stdout, "\nTesting preservation of headers...\n\n");
 		stream = g_mime_stream_file_new (stdout);
+		/*g_mime_header_set_raw (GMIME_OBJECT (message)->headers, NULL);*/
 		g_mime_header_write_to_stream (GMIME_OBJECT (message)->headers, stream);
 		g_mime_stream_flush (stream);
 		GMIME_STREAM_FILE (stream)->fp = NULL;
